"""Module for parsing a YouTube descripton into metadata tags."""
import re
from copy import deepcopy

from retag_opus import constants
from retag_opus.utils import Utils

INTERPUNCT = "\u00b7"

Tags = dict[str, list[str]]


class DescriptionParser:
    """Parse tags from YouTube description."""

    def __init__(self, manual_album_set: bool = False) -> None:
        """Create tags dictionary that will hold the parsed tags."""
        self.tags: Tags = {}
        self.base_patterns = deepcopy(constants.all_tags)
        self.performer_patterns = deepcopy(constants.performer_tags)
        self.manual_album_set = manual_album_set
        if manual_album_set:
            self.base_patterns["discsubtitle"] = self.base_patterns["album"].copy()
            self.base_patterns["discsubtitle"]["print"] = "Disc subtitle"
            self.base_patterns["album"]["pattern"] = []

    def parse_artist_and_title(self, source_line: str) -> tuple[list[str], str]:
        """Parse artist and title from standard ContentID line.

        The standard ContentID description contains an interpunct
        character separating title and the artists. This function gets
        the artist and title of the song for this line.

        :param source_line: A string containing an interpunct character.

        :return: List or artists and title of song as string.
        """
        artist_and_title = source_line.split(" " + constants.INTERPUNCT + " ")
        title = artist_and_title[0]
        artist: list[str] = artist_and_title[1:]

        if len(artist) < 2 and ", " in artist[0]:
            artist = Utils().split_tag(artist[0])

        return artist, title

    def standard_pattern(self, field_name: str, regex: str, line: str) -> None:
        """Parse metadata from line with regex and put in field_name.

        Use a regex on a standard format to extract metadata from the
        string line and set the result to the tag field_value. If the
        value already exists, remove duplicate values.
        """
        pattern = re.compile(regex)
        pattern_match = re.match(pattern, line)
        if pattern_match:
            field_value = pattern_match.groups()[len(pattern_match.groups()) - 1]
            field_value = field_value.strip()
            if self.tags.get(field_name):
                self.tags[field_name].append(field_value)
                self.tags[field_name] = Utils().remove_duplicates(self.tags[field_name])
            else:
                self.tags[field_name] = [field_value]

    def parse(self, description_tag_full: str) -> None:
        """Parse the provided youtube description.

        Parses the given description and puts the resulting tags in the
        internal tags variable of the object.

        :param description_tag_full: The full YouTube description as a
            string.
        """
        lines_since_title_artist: int = 1000

        description_tag_lines: list[str] = description_tag_full.splitlines(False)

        for description_line in description_tag_lines:
            description_line = re.sub("\n", "", description_line)
            if re.match(r"^\s*$", description_line):
                continue
            lines_since_title_artist = lines_since_title_artist + 1
            # Artist and title
            if INTERPUNCT in description_line:
                lines_since_title_artist = 0
                youtube_artist, youtube_title = self.parse_artist_and_title(description_line)
                if youtube_artist:
                    self.tags["artist"] = Utils().remove_duplicates(youtube_artist)
                self.tags["title"] = [youtube_title]

            if lines_since_title_artist == 1:
                if self.manual_album_set:
                    self.tags["discsubtitle"] = [description_line.strip()]
                else:
                    self.tags["album"] = [description_line.strip()]

            for tag_id, tag_data in self.base_patterns.items():
                for pattern in tag_data["pattern"]:
                    self.standard_pattern(tag_id, pattern, description_line)

            for tag_id, tag_data in self.performer_patterns.items():
                for pattern in tag_data["pattern"]:
                    self.standard_pattern(tag_id, pattern, description_line)

            title = self.tags.pop("title", None)
            if title:
                self.tags["title"] = [title[0]]

        artist = self.tags.get("artist")
        if artist:
            self.tags["albumartist"] = [artist[0]]
            if len(artist) > 1:
                many_artist: list[str] = []
                for a in artist:
                    many_artist = many_artist + Utils().split_tag(a)
                artist = many_artist
            else:
                artist = Utils().split_tag(artist[0])
            artist = Utils().remove_duplicates(artist)
            self.tags["artist"] = artist

        for key, value in self.tags.items():
            if value == []:
                self.tags.pop(key)

        # Basic pattern:
        # r".*[ ]   .*:\s*(.*)\s*"

        # Custom patterns
        for description_line in description_tag_lines:
            description_line = description_line.replace("\n", "")
            description_line = re.sub("\n", "", description_line)

            self.standard_pattern("copyright_date", r"\u2117 (\d\d\d\d)\s", description_line)

        copyright_date = self.tags.pop("copyright_date", None)
        date = self.tags.get("date")
        if copyright_date and not date:
            self.tags["date"] = copyright_date



"""Module containing the definition of colors used in the app."""
from typing import Final

from colorama import Fore

yt_col: Final[str] = Fore.MAGENTA
md_col: Final[str] = Fore.CYAN
man_col: Final[str] = Fore.GREEN
auto_col: Final[str] = Fore.YELLOW



"""Module for holding constants.

This module holds variables that should be constant and can be accessed
in other parts of the app.
"""
from typing import Final, TypedDict

INTERPUNCT: Final = "\u00b7"
# SPACE = "\u00b7"
SPACE: Final = " "
SEP: Final = " | "


class ParsingReference(TypedDict):
    """This is dictionary with a tag name and regex for parsing it."""

    print: str
    pattern: list[str]


tag_parse_patterns: Final[dict[str, str]] = {
    "featuring": r"(?i)(.*?)\s*[\(\[\ ](?:feat|ft|featuring)\.?\s+([^\]\)\(\[]+)[\)\]]*\s*(.*)",
    "remaster": r"(?i)(.*?)\s*[\(\[](\d{0,4}\s*remaster.*)[\)\]]\s*(.*)",
    "remaster2": r"(?i)(.*?)\s*-\s*(\d{0,4}.*remaster.*)(.*)",
    "live": r"(?i)(.*?)\s*[\(\[](live.*?)[\)\]]\s*(.*)",
    "instrumental": r"(?i)(.*?)\s*\((.*instrumental.*)\)\s*(.*)",
    "instrumental2": r"(?i)(.*?)\s*\[(.*instrumental.*)\]\s*(.*)",
    "remix": r"(?i)(.+)\s*\((.*remix.*?)\)\s*(.*)",
    "remix2": r"(?i)(.+)\s*-\s*(.*remix.*)\s*(.*)",
    "albumversion": r"(?i)(.*?)\s*[\(\[](album version.*?)[\)\]]\s*(.*)",
}

all_tags: Final[dict[str, ParsingReference]] = {
    "title": {"print": "Title", "pattern": [r".*“(.*)” by .* from ‘.*’"]},
    "album": {"print": "Album", "pattern": [r".*“.*” by .* from ‘(.*)’"]},
    "albumartist": {"print": "Album Artist", "pattern": []},
    "artist": {
        "print": "Artist(s)",
        "pattern": [
            r"^(?!.*(https?|[mM]akeup|[fF]inishing)).*[aA]rtist.*:\s*(.+)\s*",
            r".*\([fF]eat. (.+?)\)",
            r".*“.*” by (.*) from ‘.*’",
        ],
    },
    "date": {"print": "Date", "pattern": [r"Released on:\s*(\d\d\d\d-\d\d-\d\d)"]},
    "genre": {"print": "Genre", "pattern": []},
    "version": {"print": "Version", "pattern": []},
    "performer": {"print": "Performer", "pattern": [r".*[pP]erformer.*:\s*(.+)\s*"]},
    "organization": {"print": "Organization", "pattern": [r"Provided to YouTube by (.+)\s*"]},
    "copyright": {"print": "Copyright", "pattern": [r"\u2117 (.+)\s*"]},
    "composer": {"print": "Composer", "pattern": [r".*?[cC]omposer.*:\s*(.+)\s*"]},
    "conductor": {"print": "Conductor", "pattern": [r".*[cC]onductor.*:\s*(.+)\s*"]},
    "arranger": {
        "print": "Arranger",
        "pattern": [r".*?[aA]rranged\s+[bB]y.*:\s*(.+)\s*", r".*?[aA]rranger.*:\s*(.+)\s*"],
    },
    "author": {"print": "Author", "pattern": [r"(.*, )?[aA]uthor.*:\s*(.+)\s*"]},
    "producer": {"print": "Producer", "pattern": [r"(.*, )?[pP]roducer.*:\s*(.+)\s*"]},
    "publisher": {"print": "Publisher", "pattern": [r"(.*, )?[pP]ublisher.*:\s*(.+)\s*"]},
    "lyricist": {
        "print": "Lyricist",
        "pattern": [
            r"(.*, )?[wW]riter.*:\s*(.+)\s*",
            r"(.*, )?[wW]ritten\s+[bB]y.*:\s*(.+)\s*",
            r".*[lL]yricist.*:\s*(.+)\s*",
        ],
    },
}

performer_tags: Final[dict[str, ParsingReference]] = {
    "performer:vocals": {"print": "- Vocals", "pattern": [r"(.*, )?(Lead\s+)?[vV]ocal(?!.*[eE]ngineer).*:\s*(.+)\s*"]},
    "performer:background vocals": {
        "print": "- Background Vocals",
        "pattern": [r"(.*, )?[bB]ackground\s+[vV]ocal.*:\s*(.+)\s*"],
    },
    "performer:drums": {"print": "- Drums", "pattern": [r"(.*, )?[dD]rum.*:\s*(.+)\s*"]},
    "performer:percussion": {"print": "- Percussion", "pattern": [r".*[pP]ercussion.*:\s*(.+)\s*"]},
    "performer:keyboard": {"print": "- Keyboard", "pattern": [r"(.*, )?[kK]eyboard.*:\s*(.+)\s*"]},
    "performer:piano": {"print": "- Piano", "pattern": [r"(.*, )?[pP]iano.*:\s*(.+)\s*"]},
    "performer:synthesizer": {"print": "- Synthesizer", "pattern": [r".*[sS]ynth.*:\s*(.+)\s*"]},
    "performer:guitar": {
        "print": "- Guitar",
        "pattern": [r"(.*, )?[gG]uitar.*:\s*(.+)\s*" r".*[eE]lectric\s+[gG]uitar.*:\s*(.+)\s*"],
    },
    "performer:electric guitar": {"print": "- Electric guitar", "pattern": []},
    "performer:bass guitar": {"print": "- Bass guitar", "pattern": [r".*[bB]ass\s+[gG]uitar.*:\s*(.+)\s*"]},
    "performer:acoustic guitar": {"print": "- Acoustic guitar", "pattern": [r".*[aA]coustic\s+[gG]uitar.*:\s*(.+)\s*"]},
    "performer:ukulele": {"print": "- Ukulele", "pattern": [r".*[uU]kulele.*:\s*(.+)\s*"]},
    "performer:violin": {"print": "- Violin", "pattern": [r"(.*, )?[vV]iolin.*:\s*(.+)\s*"]},
    "performer:double bass": {"print": "- Double bass", "pattern": [r".*[dD]ouble\s+[bB]ass.*:\s*(.+)\s*"]},
    "performer:cello": {"print": "- Cello", "pattern": [r"(.*, )?[cC]ello.*:\s*(.+)\s*"]},
    "performer:programming": {"print": "- Programming", "pattern": [r"(.*, )?[pP]rogramm(er|ing).*:\s*(.+)\s*"]},
    "performer:saxophone": {"print": "- Saxophone", "pattern": [r"(.*, )?[sS]axophone.*:\s*(.+)\s*"]},
    "performer:flute": {"print": "- Flute", "pattern": [r"(.*, )?[fF]lute.*:\s*(.+)\s*"]},
}



"""Module for storing tags from different sources and printing them."""
import re
from copy import deepcopy
from typing import Final

from colorama import Fore
from simple_term_menu import TerminalMenu

from retag_opus import colors, constants
from retag_opus.exceptions import UserExitException
from retag_opus.utils import Utils

Tags = dict[str, list[str]]

REMOVED_TAG: Final[list[str]] = ["[Removed]"]


class MusicTags:
    """Store tags from different sources for a song.

    Stores tags for one song as a dictionary on the same format as the
    one provided by the OggOpus object.
    """

    def __init__(self, manual_album_set: bool = False) -> None:
        """Create empty attributes for each source."""
        self.base_patterns = deepcopy(constants.all_tags)
        self.performer_patterns = deepcopy(constants.performer_tags)
        if manual_album_set:
            self.base_patterns["discsubtitle"] = self.base_patterns["album"].copy()
            self.base_patterns["discsubtitle"]["print"] = "Disc subtitle"
            self.base_patterns["album"]["pattern"] = []
        self.original: Tags = {}
        self.youtube: Tags = {}
        self.fromtags: Tags = {}
        self.fromdesc: Tags = {}
        self.resolved: Tags = {}

    def print_metadata_key(
        self,
        key_type: str,
        key: str,
        key_col: str,
        data: Tags,
    ) -> None:
        """Print a single metadata key in specified color.

        :param key_type: Printed name of key.
        :param key: What key to print.
        :param key_col: Color to print the data in.
        :param data: The tags from which to get the given key.
        """
        value = constants.SEP.join(data.get(key, ["Not found"])).replace(" ", constants.SPACE)
        if self.resolved.get(key) == ["[Removed]"]:
            key_type = f"{key_type} (Removed)"
        key_col = key_col if data.get(key) else Fore.BLACK
        print("  " + key_type + ": " + key_col + value + Fore.RESET)

    def print_metadata(self, metadata: Tags, col: str) -> None:
        """Print metadata of given set of tags in specifed color.

        :param metadata: The set of tags to print.
        :param col: The color to print the tags in.
        """
        if any(t in self.performer_patterns.keys() for t in metadata.keys()):
            print("  Performers:")
            for tag_id, tag_data in self.performer_patterns.items():
                if tag_id in metadata and metadata[tag_id] is not None:
                    self.print_metadata_key(tag_data["print"], tag_id, col, metadata)
        for tag_id, tag_data in self.base_patterns.items():
            self.print_metadata_key(tag_data["print"], tag_id, col, metadata)
        print("")

    def get_tag_data(self, tag_id: str) -> list[str]:
        """Get data from all sources for given tag.

        Returns a list of all values for a specific tag from various
        sources with color depending on source.

        :param tag_id: The tag name for which data should be returned

        :return: List with strings containing the value the given tag
        has in different sources with color codes that differ between
        sources. If no source contains the tag, an empty list is
        returned.
        """
        print_data: list[str] = []
        original = self.original.get(tag_id)
        youtube = self.youtube.get(tag_id)
        fromdesc = self.fromdesc.get(tag_id)
        fromtags = self.fromtags.get(tag_id)
        if fromtags and fromtags != original:
            print_data.append(Fore.YELLOW + " | ".join(fromtags) + Fore.RESET)
        if fromdesc and fromdesc != original:
            print_data.append(Fore.GREEN + " | ".join(fromdesc) + Fore.RESET)
        if original:
            print_data.append(colors.md_col + " | ".join(original) + Fore.RESET)
        if youtube and original != youtube:
            print_data.append(colors.yt_col + " | ".join(youtube) + Fore.RESET)
        return print_data

    def print_all(self) -> None:
        """Print all tags from all sources.

        Print a block of text with all data from all sources for a given
        tag name, differentiated by color.
        """
        performer_block = []
        there_are_tags = False
        for tag_id, tag_data in self.performer_patterns.items():
            tag_all_values = self.get_tag_data(tag_id)
            if len(tag_all_values) > 0:
                there_are_tags = True
                performer_block.append(tag_data["print"] + ": " + " | ".join(tag_all_values))

        main_block = []
        for tag_id, tag_data in self.base_patterns.items():
            tag_all_values = self.get_tag_data(tag_id)
            if len(tag_all_values) > 0:
                there_are_tags = True
                main_block.append(tag_data["print"] + ": " + " | ".join(tag_all_values))
            else:
                main_block.append(tag_data["print"] + ": " + Fore.BLACK + "Not set" + Fore.RESET)

        if there_are_tags:
            if len(performer_block) > 0:
                print("Performers:")
                print("\n".join(performer_block))
            print("\n".join(main_block) + "\n")
        else:
            print(Fore.RED + "There's no data to be printed" + Fore.RESET)

    def print_youtube(self) -> None:
        """Print all tags parsed from youtube description."""
        if self.youtube:
            self.print_metadata(self.youtube, colors.yt_col)
        else:
            print(Fore.RED + "No new data parsed from description" + Fore.RESET)

    def print_original(self) -> None:
        """Print tags from original file.

        Print tags from the original file, if they are in the predefined
        set of tags.
        """
        if self.original:
            self.print_metadata(self.original, colors.md_col)
        else:
            print(Fore.RED + "There were no pre-existing tags for this file" + Fore.RESET)

    def print_from_tags(self) -> None:
        """Print all tags parsed from original tags."""
        if self.fromtags:
            self.print_metadata(self.fromtags, Fore.YELLOW)
        else:
            print(Fore.RED + "No new data parsed from tags" + Fore.RESET)

    def print_from_desc(self) -> None:
        """Print all tags parsed from YouTube description tags."""
        if self.fromdesc:
            self.print_metadata(self.fromdesc, Fore.GREEN)
        else:
            print(Fore.RED + "No new data parsed from tags parsed from description" + Fore.RESET)

    def print_resolved(self, print_all: bool = False) -> None:
        """Print resolved tags.

        Print the tags in the "resolved" set, meaning that they have
        been assumed to be correct or that they have been manually
        selected to be right.

        :param print_all: Print tags even if they haven't been changed.
        """
        other_tags: list[str] = []
        for tag_id in self.resolved.keys():
            if tag_id not in self.base_patterns and tag_id not in self.performer_patterns:
                other_tags.append(tag_id)
        if other_tags:
            print("  Other tags:")
            for tag in other_tags:
                resolved_tag = self.resolved.get(tag, [])
                joined_tag_items = " | ".join(resolved_tag).replace("\n", " ")
                if resolved_tag == REMOVED_TAG:
                    self.print_metadata_key(f"- {tag}", tag, Fore.RED, self.original)
                elif resolved_tag != self.original.get(tag):
                    self.print_metadata_key(f"- {tag}", tag, Fore.GREEN, self.resolved)
                elif len(joined_tag_items) > 200:
                    print(f"  - {tag}: {colors.md_col}[{joined_tag_items[:65]}...]{Fore.RESET}")
                else:
                    self.print_metadata_key(f"- {tag}", tag, colors.md_col, self.resolved)

        if "performer:" in " ".join(self.resolved.keys()):
            print("  Performers:")
            for tag_id, tag_data in self.performer_patterns.items():
                resolved_tag = self.resolved.get(tag_id, [])
                all_sources_tag = self.get_tag_data(tag_id)
                # If the user chose to remove a tag that existed before
                if resolved_tag == REMOVED_TAG:
                    self.print_metadata_key(tag_data["print"], tag_id, Fore.RED, self.original)
                # If the resolved tag differs from the original tag
                elif resolved_tag != self.original.get(tag_id, []) and self.resolved.get(tag_id) is not None:
                    self.print_metadata_key(tag_data["print"], tag_id, Fore.GREEN, self.resolved)
                # original and resolved are equal, but other tags exist
                elif len(all_sources_tag) > 0 and print_all:
                    print("  " + tag_data["print"] + ": " + " | ".join(all_sources_tag))
                else:
                    self.print_metadata_key(tag_data["print"], tag_id, colors.md_col, self.resolved)
        for tag_id, tag_data in self.base_patterns.items():
            resolved_tag = self.resolved.get(tag_id, [])
            all_sources_tag = self.get_tag_data(tag_id)
            if resolved_tag == REMOVED_TAG:
                self.print_metadata_key(tag_data["print"], tag_id, Fore.RED, self.original)
            elif resolved_tag != self.original.get(tag_id, []) and self.resolved.get(tag_id) is not None:
                self.print_metadata_key(tag_data["print"], tag_id, Fore.GREEN, self.resolved)
            elif len(all_sources_tag) > 0 and print_all:
                print("  " + tag_data["print"] + ": " + " | ".join(all_sources_tag))
            else:
                self.print_metadata_key(tag_data["print"], tag_id, colors.md_col, self.resolved)

        print("")

    def switch_album_to_disc_subtitle(self, manual_album_name: str) -> None:
        """Move album tag to discsubtitle tag.

        Switches the original album tag to the discsubtitle tag so that
        it will be used in the comparison for that tag, rather than
        album, which is not used when the album is set manually.

        :param manual_album_name: What to use as the new album tag.
        """
        original_album = self.original.pop("album", None)
        if original_album is not None and original_album != [manual_album_name]:
            self.original["discsubtitle"] = original_album

    def discard_upload_date(self) -> None:
        """If the date is just the upload date, discard it."""
        if self.original.get("date") and re.match(r"\d\d\d\d\d\d\d\d", self.original["date"][0]):
            self.original.pop("date", None)

    def prune_resolved_tags(self, tags_to_delete: list[str], strings_to_delete_tags_based_on: list[str]) -> None:
        """Remove tags that the user wants removed by default."""
        for tag in tags_to_delete:
            if tag in self.original:
                self.resolved[tag] = REMOVED_TAG

        for tag_id, tag_data in self.original.items():
            for bad_word in strings_to_delete_tags_based_on:
                if any(re.fullmatch(bad_word, item) for item in tag_data):
                    self.resolved[tag_id] = REMOVED_TAG

    def add_source_tag(self) -> None:
        """Add a comment tag with the value 'youtube-dl'."""
        if any("youtube" in item for item in self.original.get("purl", [])):
            if "youtube-dl" not in self.original.get("comment", []):
                self.resolved["comment"] = self.original.get("comment", []) + ["youtube-dl"]

    def get_field(self, field: str, only_new: bool = False) -> list[str]:
        """Get values for field from all sources.

        :param field: Tag to get values for.
        :param only_new: Whether to include the value of the field in
            the original music file.
        """
        old_value = self.original.get(field, [])
        yt_value = self.youtube.get(field, [])
        from_desc_value = self.fromdesc.get(field, [])
        from_tags_value = self.fromtags.get(field, [])
        if only_new:
            return Utils().remove_duplicates(yt_value + from_desc_value + from_tags_value)
        else:
            return Utils().remove_duplicates(old_value + yt_value + from_desc_value + from_tags_value)

    def modify_resolved_field(self) -> None:
        """Manually add key-value pairs to the set of resolved tags."""
        key = " "
        val = " "
        while key and val:
            print("Enter key and value (newline cancels, sequence ' | ' splits tag):")
            key = input("  Key: ")
            val = input("  Value: ")
            if key and val:
                self.resolved[key] = val.split(" | ")

    def delete_tag_item(self) -> None:
        """Present a menu for removing tags and remove selected ones.

        Presents the user with a prompt to fill in a field name from
        which to remove keys. The user can then select what values to
        remove from the list of values from this key.
        """
        tags_in_resolved = []
        for tag in self.resolved.keys():
            if tag in self.base_patterns.keys() or tag in self.performer_patterns.keys():
                if self.resolved[tag] != REMOVED_TAG:
                    tags_in_resolved.append(tag)

        tags_in_resolved = sorted(tags_in_resolved)
        tags_in_resolved.append("Quit")
        removal_menu = TerminalMenu(tags_in_resolved, title="Which field do you want to delete items from?")
        tag_index = removal_menu.show()

        if tag_index is None or isinstance(tag_index, tuple) or tags_in_resolved[tag_index] == "Quit":
            print(Fore.YELLOW + "Returning without removing anything" + Fore.RESET)
        else:
            selected_tag = tags_in_resolved[tag_index]
            items_in_tag = self.resolved.get(selected_tag, []).copy()
            item_removal_menu = TerminalMenu(
                items_in_tag,
                title=f"Which field should be removed from the '{selected_tag}' tag?",
                multi_select=True,
                show_multi_select_hint=True,
                multi_select_empty_ok=True,
                multi_select_select_on_accept=False,
            )

            items_to_remove = item_removal_menu.show()
            if isinstance(items_to_remove, int):
                items_to_remove = [items_to_remove]
            elif items_to_remove is None or len(items_to_remove) == 0:
                print(Fore.YELLOW + "Returning without removing anything" + Fore.RESET)
                return

            for item in items_to_remove:
                self.resolved[selected_tag].remove(items_in_tag[item])
                if len(self.resolved[selected_tag]) == 0:
                    self.resolved[selected_tag] = REMOVED_TAG

    def check_any_new_data_exists(self) -> bool:
        """Check whether there are new tags in the sources.

        Check whether the parsing has produced any values for tags that
        weren't already in the metadata of the music file.

        :return: True if there are new values found, otherwise False.
        """
        new_content_exists = False
        all_new_fields = [key for key in self.youtube.keys()]
        all_new_fields += [key for key in self.fromdesc.keys()]
        all_new_fields += [key for key in self.fromtags.keys()]
        all_new_fields = Utils.remove_duplicates(all_new_fields)
        for tag in all_new_fields:
            all_sources = set(self.get_field(tag))
            original_tags = set(self.original.get(tag, []))
            if not original_tags.issuperset(all_sources):
                # If there are no new tags, the set of all tags should
                # not contain anything that is not already in the
                # original tags.
                new_content_exists = True

        for tag in self.resolved.keys():
            original_tags = set(self.original.get(tag, []))
            resolved_tags = set(self.resolved.get(tag, []))
            if not original_tags.issuperset(resolved_tags):
                # There may be some automatically added tags in
                # resolved. Check that resolved doesn't contain any tag
                # that doesn't already exist in the original tags.
                new_content_exists = True

        return new_content_exists

    def determine_album_artist(self) -> None:
        """Choose value to use for albumartist tag."""
        all_artists = self.get_field("artist")
        resolved_artist = self.resolved.get("artist")
        original_artist = self.original.get("artist")

        if len(all_artists) > 1:
            print("-----------------------------------------------")
            self.print_resolved(print_all=True)
            print(Fore.BLUE + "Select the album artist:")
            one_artist = Utils().select_single_tag(all_artists)
            if len(one_artist) > 0:
                self.resolved["albumartist"] = one_artist
        elif not self.original.get("albumartist"):
            if resolved_artist:
                self.resolved["albumartist"] = resolved_artist
            elif original_artist:
                self.resolved["albumartist"] = original_artist

    def default_to_youtube_date(self) -> None:
        """If youtube has date data, use that for resolved."""
        date = self.youtube.get("date")
        if date:
            self.resolved["date"] = date

    def set_artist_if_obvious(self) -> None:
        """Set the resolved artist if it's good for sure.

        If the original tag and the youtube tags are identical
        with regards to the artist tag, assume that it is correct.
        """
        md_artist = self.original.get("artist")
        yt_artist = self.youtube.get("artist")
        if md_artist is not None and yt_artist is not None:
            if (
                len(md_artist) == 1
                and Utils.split_tag(md_artist[0]) == yt_artist
                or Utils.is_equal_when_stripped(md_artist, yt_artist)
            ):
                self.resolved["artist"] = yt_artist

    def manually_adjust_tag_when_resolving(self, tag_name: str) -> bool:
        """Perform a manual action on a tag when no source is right.

        For the tag with the given name, manually fill choose its
        content. Either completely manually or with help from the
        found tags.

        When neither the original tag, nor any of the new ones has the
        right value, the user can remove the whole tag, one or more
        values in it, a specific value from the found ones, or just
        manually write the tag.
        """
        other_choices = [
            "[s] Select items from a list",
            "[m] Manually fill in tag",
            "[p] Print description metadata",
            "[r] Remove field",
            "[g] Go back",
        ]

        other_choice_menu = TerminalMenu(other_choices, title="Choose the source you want to use:")
        choice = other_choice_menu.show()

        action = ""
        if isinstance(choice, int):
            action = other_choices[choice]

        match action:
            case "[m] Manually fill in tag":
                manual_tag = input("Value: ")
                if manual_tag:
                    self.resolved[tag_name] = [manual_tag]
                else:
                    print("No input, not changing tag")
                    return True
                return False
            case "[p] Print description metadata":
                print("-----------------------------------------------")
                self.print_youtube()
                return True
            case "[s] Select items from a list":
                available_tags = self.get_field(tag_name)
                tag_selection_menu = TerminalMenu(
                    available_tags,
                    title="Select the items you want in this tag",
                    multi_select=True,
                    show_multi_select_hint=True,
                    multi_select_empty_ok=True,
                    multi_select_select_on_accept=False,
                )

                items = tag_selection_menu.show()
                if isinstance(items, int):
                    items = [items]
                elif items is None:
                    print(Fore.RED + "Invalid choice, try again")
                    return True

                self.resolved[tag_name] = []
                for item in items:
                    self.resolved[tag_name].append(available_tags[item])

                return False
            case "[r] Remove field":
                self.resolved[tag_name] = REMOVED_TAG
                return False
            case _:
                print("Going back to previous menu")
                return True

    def resolve_metadata(self) -> None:
        """Merge the metadata from the different sources.

        Use the acquired metadata from all sources to produce a set of
        "resolved" tags. A few tags will automatically be used, but most
        require user interaction to resolve. The user is presented with
        menus where the selection happens. It's also possible to make
        manual edits to the tags.

        :raises UserExitException: Raised when the user chooses to quit
            the app.
        """
        all_tags_with_new_data = [tag_name for tag_name in self.youtube.keys()]
        all_tags_with_new_data += [tag_name for tag_name in self.fromdesc.keys()]
        all_tags_with_new_data += [tag_name for tag_name in self.fromtags.keys()]
        all_tags_with_new_data = Utils.remove_duplicates(all_tags_with_new_data)

        for tag_name in all_tags_with_new_data:
            old_value = self.original.get(tag_name, [])
            yt_value = self.youtube.get(tag_name, [])
            from_desc_value = self.fromdesc.get(tag_name, [])
            from_tags_value = self.fromtags.get(tag_name, [])
            all_values = self.get_field(tag_name, only_new=True)

            if not old_value and len(all_values) > 0 and tag_name != "albumartist":
                if len(yt_value) > 0:
                    self.resolved[tag_name] = yt_value
                elif len(from_tags_value) > 0:
                    self.resolved[tag_name] = from_tags_value
                else:  # len(from_desc_value) > 0:
                    self.resolved[tag_name] = from_desc_value
                print(
                    Fore.YELLOW + f"{tag_name.title()}: No value exists in metadata. Using parsed data: "
                    f"{self.resolved[tag_name]}." + Fore.RESET
                )
            elif Utils.is_equal_when_stripped(yt_value, old_value) and len(old_value) > 0:
                print(Fore.GREEN + f"{tag_name.title()}: Metadata matches YouTube description tags." + Fore.RESET)
                self.resolved[tag_name] = [v.strip() for v in old_value]
                continue
            elif Utils.is_equal_when_stripped(from_desc_value, old_value) and len(old_value) > 0:
                print(Fore.GREEN + f"{tag_name.title()}: Metadata matches tags parsed from YouTube tags." + Fore.RESET)
                self.resolved[tag_name] = [v.strip() for v in old_value]
                continue
            elif Utils.is_equal_when_stripped(from_tags_value, old_value) and len(old_value) > 0:
                print(Fore.GREEN + f"{tag_name.title()}: Metadata matches tags parsed from original tags." + Fore.RESET)
                self.resolved[tag_name] = [v.strip() for v in old_value]
                continue
            else:
                redo = True
                print("-----------------------------------------------")
                self.print_resolved(print_all=True)
                while redo:
                    redo = False
                    candidates = []
                    print(
                        Fore.RED
                        + f"{tag_name.title()}: Mismatch between values in description and metadata:"
                        + Fore.RESET
                    )
                    if len(yt_value) > 0:
                        print("YouTube description: " + colors.yt_col + " | ".join(yt_value) + Fore.RESET)
                        candidates.append("YouTube description")
                    if len(from_tags_value) > 0:
                        print("Parsed from original tags: " + Fore.YELLOW + " | ".join(from_tags_value) + Fore.RESET)
                        candidates.append("Parsed from original tags")

                    print("Exisiting metadata:  " + colors.md_col + " | ".join(old_value) + Fore.RESET)
                    candidates.append("Existing metadata")

                    if len(from_desc_value) > 0:
                        print("Parsed from YouTube tags: " + Fore.GREEN + " | ".join(from_desc_value) + Fore.RESET)
                        candidates.append("Parsed from Youtube tags")

                    candidates.append("Other action")
                    candidates.append("Quit")

                    candidate_menu = TerminalMenu(candidates)
                    choice = candidate_menu.show()

                    if not isinstance(choice, int):
                        raise UserExitException("Skipping this and all later songs")

                    match candidates[choice]:
                        case "Other action":
                            redo = self.manually_adjust_tag_when_resolving(tag_name)

                        case "Existing metadata":
                            self.resolved[tag_name] = old_value

                        case "YouTube description":
                            self.resolved[tag_name] = yt_value

                        case "Parsed from original tags":
                            self.resolved[tag_name] = from_tags_value

                        case "Parsed from Youtube tags":
                            self.resolved[tag_name] = from_desc_value

                        case _:
                            raise UserExitException("Skipping this and all later songs")

        self.determine_album_artist()



"""Module with helper functions for the rest of the app."""
import re
import sys
from pathlib import Path
from typing import List

from colorama import Fore
from simple_term_menu import TerminalMenu

from retag_opus import constants


class Utils:
    """Class containing static utility functions."""

    @staticmethod
    def remove_duplicates(duplicates: list[str]) -> list[str]:
        """Remove duplicate elemets from given list."""
        return list(dict.fromkeys(duplicates))

    @staticmethod
    def prune_title(original_title: str) -> str:
        """Clean up title tag value.

        Remove things such as live version specification and featured
        artist information from the title of a song, and remove leading
        and trailing whitespace.
        """
        pruned = original_title
        for pattern in constants.tag_parse_patterns.values():
            pruned = re.sub(pattern, r"\1 \3", pruned)
        return pruned.strip()

    @staticmethod
    def split_tag(input: str) -> List[str]:
        """Split the provided tag by pre-defined delimeters."""
        split_tags = re.split(", | and | & |; ", input)
        split_pruned_tags = [t.strip() for t in split_tags]
        return split_pruned_tags

    @staticmethod
    def file_path_to_song_data(file_path: Path) -> str:
        """Convert file name to readable string.

        Produce a string with song title and artist on a readable format
        from the name of the file being worked with.
        """
        text_to_print = str(file_path)
        basename_match = re.match(r"(.+/)*(.*).opus", text_to_print)
        if basename_match:
            basename = basename_match.groups()[1]
            text_to_print = basename
            playlist_match = re.match(r"<(.*)> - <(.*?)(?: - Topic)*> - <(.*)> - <(.*)>", text_to_print)
            artist_title_match = re.match("<(.*?)(?: - Topic)*> - <(.*)>", text_to_print)
            title_only_match = re.match("<(.*)>", text_to_print)
            if playlist_match:
                text_to_print = playlist_match.groups()[1] + " - " + playlist_match.groups()[2]
            elif artist_title_match:
                text_to_print = artist_title_match.groups()[0] + " - " + artist_title_match.groups()[1]
            elif title_only_match:
                text_to_print = title_only_match.groups()[0]
        return text_to_print

    @staticmethod
    def select_single_tag(candidates: list[str]) -> list[str]:
        """Select a single value to keep for a tag from a menu.

        Let's the user select one out of the candidate tags and returns
        a list with that as the only item. The user can also choose to
        skip with q, escape, or choosing the --No change-- item.

        :param candidates: The list of tags to choose between.

        :return: A list with the selected candidate or an empty list if
        none is selected.
        """
        candidates.append("--No change--")
        choose_one_tag_menu = TerminalMenu(candidates, title="Choose one tag to use")
        choice = choose_one_tag_menu.show()
        if choice is None:
            print(Fore.YELLOW + "No tag selected")
        elif isinstance(choice, int) and choice != len(candidates) - 1:
            print(Fore.BLUE + f"Using {candidates[choice]} as tag for this song")
            return [candidates[choice]]
        return []

    @staticmethod
    def exit_now() -> None:
        """Immediately exit the program."""
        sys.exit(0)

    @staticmethod
    def is_equal_when_stripped(values_1: list[str], values_2: list[str]) -> bool:
        """Compare lists disregarding whitespace.

        Check if content of lists of strings are the same after strings
        have been stripped.
        """
        cleaned_values_1 = [v.strip() for v in values_1]
        cleaned_values_2 = [v.strip() for v in values_2]
        return sorted(cleaned_values_1) == sorted(cleaned_values_2)



"""Module for parsing text in the existing metadata keys.

This module is for parsing text in the existing metadata keys, rather
that the YouTube description.
"""
import re
from typing import Dict, Final, List

from retag_opus import constants
from retag_opus.utils import Utils

INTERPUNCT: Final[str] = "\u00b7"


class TagsParser:
    """Clean and parse already existing sets of tags.

    This class is for working with a set of tags and parsing metadata
    from them, such as the version of a song (e.g. "live version"), but
    also for cleaning up the tags by e.g. splitting them by pre-defined
    delimeters.
    """

    def __init__(self, tags: Dict[str, List[str]]):
        """Set attributes of the TagsParser."""
        self.tags: Dict[str, List[str]] = {}
        self.original_tags = tags

    def parse_tags(self) -> None:
        """Look through old tags for metadata.

        Go through some of the existing tags and see if they contain
        information that should go into another tag.
        """
        old_title = self.original_tags.get("title", [])

        old_artist = self.original_tags.get("artist", [])
        if len(old_artist) == 1:
            old_artist = Utils.split_tag(old_artist[0])
        new_artist = []

        old_version = self.original_tags.get("version", [])
        new_version = []

        old_genre = self.original_tags.get("genre", [])
        new_genre = []

        for title in old_title:
            featuring_regex = constants.tag_parse_patterns["featuring"]
            pattern_match = re.match(featuring_regex, title)
            if pattern_match:
                new_artist += Utils().split_tag(pattern_match.groups()[1].strip())

            live_regex = constants.tag_parse_patterns["live"]
            live_match = re.match(live_regex, title)
            if live_match:
                new_version.append(live_match.groups()[1].strip())

            albumversion_regex = constants.tag_parse_patterns["albumversion"]
            albumversion_match = re.match(albumversion_regex, title)
            if albumversion_match:
                new_version.append(albumversion_match.groups()[1].strip())

            instrumental_regex = constants.tag_parse_patterns["instrumental"]
            instrumental_match = re.match(instrumental_regex, title)
            if instrumental_match:
                new_genre.append("Instrumental")

            instrumental2_regex = constants.tag_parse_patterns["instrumental2"]
            instrumental2_match = re.match(instrumental2_regex, title)
            if instrumental2_match:
                new_genre.append("Instrumental")

            remix_regex = constants.tag_parse_patterns["remix"]
            remix_match = re.match(remix_regex, title)
            if remix_match:
                new_version.append(remix_match.groups()[1].strip())

            remix2_regex = constants.tag_parse_patterns["remix2"]
            remix2_match = re.match(remix2_regex, title)
            if remix2_match:
                new_version.append(remix2_match.groups()[1].strip())

            remaster_regex = constants.tag_parse_patterns["remaster"]
            remaster_match = re.match(remaster_regex, title)
            if remaster_match:
                new_version.append(remaster_match.groups()[1].strip())

            remaster2_regex = constants.tag_parse_patterns["remaster2"]
            remaster2_match = re.match(remaster2_regex, title)
            if remaster2_match:
                new_version.append(remaster2_match.groups()[1].strip())

        if set(new_version) != set(old_version) and len(new_version) > 0:
            self.tags["version"] = Utils().remove_duplicates(old_version + new_version)

        if set(new_artist) != set(old_artist) and len(new_artist) > 0:
            self.tags["artist"] = Utils().remove_duplicates(old_artist + new_artist)

        if set(new_genre) != set(old_genre) and len(new_genre) > 0:
            self.tags["genre"] = Utils().remove_duplicates(old_genre + new_genre)

        if len(old_title) > 0:
            pruned_title = Utils().prune_title(old_title[0])
            if old_title[0] != pruned_title:
                self.tags["title"] = [pruned_title]

    def split_select_original_tags(self) -> None:
        """Split certain original tags on predefined separators.

        Split the genre and artist tag into a list if they contains
        pre-defined separators. New tags are only set if there is
        anything that differs from the old tags.
        """
        tags_to_split = ["genre", "artist"]

        for tag in tags_to_split:
            tags_tag = self.original_tags.get(tag)
            if tags_tag is not None and not len(tags_tag) > 1:
                new_tag = Utils().split_tag(tags_tag[0])
                if new_tag != tags_tag:
                    self.tags[tag] = new_tag



"""Module for parsing command line arguments."""
import argparse
from argparse import Namespace
from typing import Sequence

import shtab

from retag_opus import __version__


class Cli:
    """Class for parsing command line arguments."""

    @staticmethod
    def parse_arguments(argv: Sequence[str] | None) -> Namespace:
        """Create parser and parse CLI arguments, and return them."""
        parser = argparse.ArgumentParser()

        shtab.add_argument_to(parser, ["-s", "--print-completion"])

        parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            default=False,
            dest="fix_descriptionless",
            help="Even if there is no YouTube description, suggest improving existing tags",
        )

        parser.add_argument(
            "-b",
            "--album",
            action="store",
            required=False,
            default=None,
            dest="manual_album",
            help="Manually sets the album tag to the given value and puts any parsed album in the " "discsubtitle tag",
        )

        parser.add_argument(
            "-d",
            "--directory",
            action="store",
            required=True,
            dest="dir",
            help="directory in which the files to be retagged are " "located",
        ).complete = shtab.DIRECTORY  # type: ignore

        parser.add_argument(
            "-V",
            "--version",
            action="version",
            version=f"retag (version {__version__})",
        )

        return parser.parse_args(argv)



"""An app to transfer YouTube data about a song to the song metadata.

This in ann app that parses a "description" tag in a song, corresponding
to the youtube version of the song, and parse it to produce other tags.
Existing tags are also used to produce new ones, e.g. moving strings
like "2020 Remix" from the title tag to the version tag.
"""

import os
import tomllib
from copy import deepcopy
from pathlib import Path
from typing import Sequence

from colorama import Fore, init
from mutagen.oggopus import OggOpus
from simple_term_menu import TerminalMenu

from retag_opus import colors
from retag_opus.cli import Cli
from retag_opus.description_parser import DescriptionParser
from retag_opus.exceptions import UserExitException
from retag_opus.music_tags import REMOVED_TAG, MusicTags
from retag_opus.tags_parser import TagsParser
from retag_opus.utils import Utils

init(autoreset=True)

Tags = dict[str, list[str]]

CONFIG_DIR = Path(os.environ.get("XDG_CONFIG_DIR", Path.home() / ".config"))
CONFIG_PATH = CONFIG_DIR / "retag.toml"


def run(argv: Sequence[str] | None = None) -> int:
    """Run all the functionality of the app."""
    args = Cli.parse_arguments(argv)
    music_dir = Path(args.dir).resolve()
    try:
        with open(CONFIG_PATH, "rb") as f:
            config = tomllib.load(f)
    except FileNotFoundError:
        config = {}
    if not music_dir.is_dir():
        print(Fore.RED + f"{args.dir} is not a directory!")
        return 1

    all_files = list(filter(Path.is_file, Path(music_dir).glob("*.opus")))
    if not all_files:
        print(Fore.YELLOW + f"There appears to be no .opus files in the provided directory {args.dir}")
        return 0

    for idx, file_path in enumerate(all_files):
        redo = True
        file_name = Utils().file_path_to_song_data(file_path)
        while redo:
            redo = False
            # Print info about file and progress
            print("\n" + Fore.BLUE + f"Song {idx + 1} of {len(all_files)}" + Fore.RESET)
            print(Fore.BLUE + f"----- Song: {file_name} -----" + Fore.RESET)

            # 1. Read the data and make basic improvements
            old_metadata: OggOpus = OggOpus(file_path)  # type: ignore
            old_tags: Tags = {}
            tag_source = {}
            for key, val in old_metadata.items():  # type: ignore
                old_tags[key] = val
            for key in old_tags.keys():
                tag_source[key] = colors.md_col
            tags = MusicTags(manual_album_set=args.manual_album is not None)

            tags.original = old_tags
            tags.discard_upload_date()
            if args.manual_album is not None:
                tags.switch_album_to_disc_subtitle(args.manual_album)

            tags.resolved = deepcopy(old_tags)

            old_tags_parser = TagsParser(tags.original)
            old_tags_parser.parse_tags()
            old_tags_parser.split_select_original_tags()
            tags.fromtags = old_tags_parser.tags

            # 1.1 Manually set album
            if args.manual_album is not None:
                tags.resolved["album"] = [args.manual_album]

            # 2. Get description
            description_lines: list[str] | None = old_tags.get("synopsis")
            if description_lines is None:
                description_lines = old_tags.get("description")

            # 3. If description exists, send it to be parsed
            if description_lines:
                desc_parser = DescriptionParser(manual_album_set=args.manual_album is not None)
                description = "\n".join(description_lines)
                desc_parser.parse(description)
                tags.youtube = desc_parser.tags
                tags.add_source_tag()

                new_tags_parser = TagsParser(tags.youtube)
                new_tags_parser.parse_tags()
                tags.fromdesc = new_tags_parser.tags

            # 4.5 Get rid of shady tags
            tags_to_delete = config.get("tags_to_delete", [])
            strings_to_delete_tags_based_on = config.get("strings_to_delete_tags_based_on", [])
            tags.prune_resolved_tags(tags_to_delete, strings_to_delete_tags_based_on)

            if not tags.check_any_new_data_exists() and not args.manual_album:
                print(Fore.YELLOW + "No new data exists. Skipping song." + Fore.RESET)
                break

            # 4. For each field, if there are conflicts, ask user input
            try:
                tags.resolve_metadata()
            except UserExitException as e:
                print(f"RetagOpus exited successfully: {e}")
                return 0

            # 5. Show user final result and ask if it should be saved or
            # retried, or song skipped
            reshow_choices = True

            while reshow_choices:
                print("Final result:")
                tags.print_resolved()
                reshow_choices = False
                options = [
                    "[p] pass",
                    "[s] save",
                    "[r] reset",
                    "[m] modify tag",
                    "[d] delete item in tag",
                    "[y] youtube description",
                    "[a] all metadata",
                    "[e] resolved metadata",
                    "[q] quit",
                ]
                terminal_menu = TerminalMenu(options, title="What do you want to do?")
                choice = terminal_menu.show()
                print("-" * 40)

                action = "[q] quit"
                if choice is not None and not isinstance(choice, tuple):
                    action = options[choice]

                match action:
                    case "[q] quit":
                        print("RetagOpus exited successfully: Skipping this and all later songs")
                        return 0
                    case "[s] save":
                        for tag, data in tags.resolved.items():
                            if data == REMOVED_TAG:
                                old_metadata.pop(tag, None)  # type: ignore
                            else:
                                old_metadata[tag] = data
                        old_metadata.save()
                        print(Fore.GREEN + f"Metadata saved for file: {file_name}")
                    case "[r] reset":
                        print(f"Trying to improve metadata again for file: {file_name}")
                        redo = True
                    case "[m] modify tag":
                        tags.modify_resolved_field()
                        print(Fore.BLUE + "Current metadata to save:")
                        tags.print_resolved()
                        reshow_choices = True
                    case "[d] delete item in tag":
                        tags.delete_tag_item()
                        reshow_choices = True
                    case "[p] pass":
                        print(Fore.YELLOW + f"Pass. Skipping song: {file_name}")
                    case "[y] youtube description":
                        if description_lines:
                            print(Fore.BLUE + "Original YouTube description:")
                            print(colors.yt_col + "\n".join(description_lines))
                        else:
                            print(Fore.RED + "No YouTube description tag for this song.")
                        reshow_choices = True
                    case "[a] all metadata":
                        print(Fore.BLUE + "All old and new metadata suggested for this file:")
                        tags.print_all()
                        reshow_choices = True
                    case "[e] resolved metadata":
                        print(Fore.BLUE + "Current metadata to save:")
                        tags.print_resolved()
                        reshow_choices = True
                    case _:
                        print(Fore.RED + "Something went wrong, starting over")
                        redo = True

    return 0
